#include "etherx.h"


/*
 *  This is the callback for the IO loop
 *  -p is a null terminated string read from the c connection
 *  flag is to signal special conditions
 */
conn *dispatch(char *p, conn *c, int flag)
{
	conn *c_new = NULL;
	int s;
	transport *t;
	xpt *xml, *next;
	jpair *par;

	DBG_VAL("dispatch_data",p);

	switch(flag)
	{
		case IO_REMOVE:
			DBG_VAL("removing_connection", c->name);
			transport_del(transport_get(c));
			break;

		case IO_INIT:
			DBG_MSG("Initialization from IO loop");

			/* need to create listen sockets on localhost
			 * and network ports
			 */
			s = make_netsocket(DEFAULT_PORT, jpair_getval(etc.vars, "local"), 0);
			if(s < 0)
			{
				ERROR("main", "Could not create localhost socket, exiting");
				sig_die();
			}
			if(listen(s, 5) < 0)
			{
				ERROR("main", "Could not listen on localhost socket, exiting");
				sig_die();
			}
			DBG_VAL("listening_on", jpair_getval(etc.vars, "local"));
			c_new = conn_add(c_new, s);
			c_new->type = CONN_LISTEN;
			c_new->name = strdup("local");

			par = etc.vars;
			while(par != NULL)
			{
				if(par->key != NULL && strcmp(par->key, "listen") == 0)
				{
					s = make_netsocket(DEFAULT_PORT, par->val, 0);
					if(s < 0)
					{
						syslog_debug(SYSLOG_ERROR, "main", "new_socket_failed", par->val);
						par = par->next;
						continue;
					}
					if(listen(s, 5) < 0)
					{
						syslog_debug(SYSLOG_ERROR, "main", "new_socket_listen_failed", par->val);
						par = par->next;
						continue;
					}
					c_new = conn_add(c_new, s);
					c_new->type = CONN_LISTEN;
					c_new->name = strdup("listen");
					DBG_VAL("listening_on", par->val);
				}
				par = par->next;
			}
			break;
		case IO_NORMAL:
			DBG_MSG("Processing normally");
			if(c->xp == NULL) /* initialize xp if it hasn't been yet */
				c->xp = xpt_packet_new();

			/* process the string for any xml packets */
			xpt_packet_data(c->xp,p,strlen(p));

			/* loop through and handle any new packets found */
			next = xml = c->xp->packets;
			while(next != NULL)
			{
				xml = next;
				next = next->next;
				xml->next = NULL; /* single it out */
				handle_packet(c, xml);
			}
			free_xpt(c->xp->packets);
			c->xp->packets = NULL;

			/* uhoh!  XML error! */
			if(c->xp->flag != XPT_POOLFLAG_NONE)
			{
				switch(c->xp->flag)
				{
					case XPT_POOLFLAG_ERROR:
						DBG_VAL("parsing_error_from",c->name);
						break;
					case XPT_POOLFLAG_CLOSE:
						DBG_VAL("close_from",c->name);
						break;
				}
				DBG_VAL("removing_connection", c->name);
				transport_del(transport_get(c));
			}
			break;
	}

	/* final check, if any new transports have been connected and we need to let the io loop know */
	t = etc.transports;
	while(t != NULL)
	{
		if(t->flag_new == 1)
		{
			t->flag_new = 0;
			return t->c;
		}
		t = t->next;
	}

	return c_new;
}

void handle_packet(conn *c, xpt *x)
{
	jpair *par, *flipa, *flipb;
	transport *to, *t;
	char *str, *str2;

	t = transport_get(c);
	if(t == NULL) /* data from new connection, add into lookup table */
	{
		if(c->xp->name == NULL || strcmp(c->xp->name,"ether") != 0)
		{
			/* a valid <ether> header wasn't sent */
			DBG_VAL("illegal_data",c->name);
			return;
		}else{
			DBG_VAL("new_transport",c->name);
			transport_add(c, 0);
			t = etc.transports;
		}
	}
	time(t->activity);

	par = xpt2jpair(x);

	if(par == NULL || par->val == NULL)
		return;
	DBG_VAL("packet",par->key);

	if(t->type == TRANSPORT_LOCAL)
	{
		/* check for and handle the only special packet, handshake */
		if(strcmp(par->key,"alias") == 0)
		{
			DBG_VAL("add_alias",par->val);
			lookup_add(par->val, t);
			if(jpair_getval(x->attributes,"default") != NULL)
				c->name = strdup(par->val);

			free_jpair(par);
			return;
		}

		if(strcmp(par->key,"handshake") == 0)
		{
			/* pass this transport it's main config file section in return */
			DBG_VAL("get_config",par->key);
			str = strgrow(NULL,"<config>",0,0);
			str = strgrow(str,jpair_getval(etc.transport_config,par->val),1,0);
			str = strgrow(str,"</config>",1,0);

			str2 = create_packet("config",NULL,str);
			c->packets = packet_add(c->packets, str2, NULL);

			free(str);
			free(str2);
			free_jpair(par);
			return;
		}
	}

	if(jpair_get(x->attributes, "to") == NULL || jpair_get(x->attributes, "from") == NULL)
		return;

	/* this is where the packet gets routed or bounced */
	to = resolve(jpair_getval(x->attributes, "to"));
	/* only local transports can ask for new connections */
	if(to == NULL && t->type == TRANSPORT_LOCAL)
		to = transport_connect(jpair_getval(x->attributes, "to"));
	if(to == NULL || (t->type == TRANSPORT_NET && to->type == TRANSPORT_NET) || t == to)
	{
		/* right now this means either DNS failed, there's no transport for that name, or security won't allow two net-sockets to talk to each other... we could be more specific here :) */
		DBG_MSG("Packet delivery failed, return to sender");
		x->attributes = jpair_new(x->attributes,"error_type","Server Name Resolution Failed", 0);
		x->attributes = jpair_new(x->attributes,"error_packet",x->name, 0);

		/* flip from/to */
		flipa = jpair_get(x->attributes,"from");
		flipb = jpair_get(x->attributes,"to");
		free(flipa->key);
		free(flipb->key);
		flipa->key = strdup("to");
		flipb->key = strdup("from");

		free(par->key);
		par->key = strdup("error");
		to = t;
	}

	str = create_packet(par->key, x->attributes, par->val);
	to->c->packets = packet_add(to->c->packets, str, NULL);
	free(str);

	free_jpair(par);
}

char *create_packet(char *name, jpair *atts, char *body)
{
	xpt *next;
	char *ret = NULL;
	char *buf = NULL;

	ret = strgrow(ret,"<",1,0);
	ret = strgrow(ret,name,1,0);
	while(atts != NULL)
	{
		ret = strgrow(ret," ",1,0);
		ret = strgrow(ret,atts->key,1,0);
		ret = strgrow(ret,"='",1,0);
		ret = strgrow(ret,strescape(atts->val),1,1);
		ret = strgrow(ret,"'",1,0);
		atts = atts->next;
	}
	ret = strgrow(ret,"><![CDATA[",1,0);
	ret = strgrow(ret,body,1,0);
	ret = strgrow(ret,"]]></",1,0);
	ret = strgrow(ret,name,1,0);
	ret = strgrow(ret,">",1,0);

	return ret;
}
