#include "common.h"

#include "packet.h"
#include "conn.h"

/* io.c
 *
 * This file contains the main IO loop.  Accepts incoming connections and listens on
 * all sockets for data, when data is found it passes it to a callback function in the main app.
 *
 */

jmp_buf _jabberlib_ioenv;
int _jabberlib_dbg;
void (*_jabberlib_logger)(int level, char *from, char *message);

/* some crap to be removed when error checking is done correctly */
void h_err(int check, char *err)
{
	if(check < 0)
	{
		syslog(SYSLOG_ERROR, err, strerror(errno));
		raise(SIGTERM);
	}
}

void collect_data(conn *(*handle_data)())
{
	char buffer[MAXMSG];
	int new, nbytes, err, maxfd = 0, first = 1;
	fd_set active_fd_set, read_fd_set, write_fd_set;
	struct sockaddr_in sa;
	size_t size = sizeof(sa);
	conn *c_new, *c, *c_count;
	packet *p;

	c_new = c = c_count = NULL;

	DBG_MSG("IO: collecting data");

	FD_ZERO(&active_fd_set);
	FD_ZERO(&write_fd_set);

	err = setjmp(_jabberlib_ioenv);

	while(err == 0)
	{
		read_fd_set = active_fd_set;

		/* first time through we initialize the IO routine */
		if(!first)
		{
			DBG_MSG("IO: waiting");
			h_err(select(maxfd + 1, &read_fd_set, &write_fd_set, NULL, NULL),"select");

			FD_ZERO(&active_fd_set);
		}else{
			c = (*handle_data)(NULL,NULL,IO_INIT);
			first = 0;
		}

		/*
		 * Loop through all of the connections to read
		 */
		c_count = c;
		while(c_count != NULL)
		{
			if(FD_ISSET(c_count->id, &read_fd_set))
			{
		 		if(c_count->type == CONN_LISTEN)
				{ /* Accept new connections from a listen socket */
					/* put a while loop here?  can there be multiple connections waiting? */
					new = accept(c_count->id,(struct sockaddr *) &sa, &size);
					h_err(new,"accept");
					DBG_VAL("new_connection",inet_ntoa(sa.sin_addr));
					/*printf("Connecting socket [%d] from host [%s], source port [%hd]\n",new,inet_ntoa(sa.sin_addr),ntohs(sa.sin_port));*/
					c = conn_add(c,new);
					c->name = strdup(inet_ntoa(sa.sin_addr));
					c->ver = strdup(c_count->name); /* express the parent socket that it was created from */
					if(new > maxfd)
						maxfd = new;
					FD_SET(new, &active_fd_set);
				}else{
					DBG_VAL("new_data_from",c_count->name);
					nbytes = 0;
					if(c_count->type != CONN_RAW)
					{
						nbytes = read(c_count->id, buffer, MAXMSG-1);
		
						/* Submitted by sperrine@xuma.com */
						if (errno != ECONNRESET && errno != EPIPE)
							h_err(nbytes,"read");

						if(nbytes <= 0)
						{
							DBG_VAL("dropping_connection",c_count->name);
							FD_CLR(c_count->id, &active_fd_set); 	/* remove it from the select */
							c_new = c_count->next;	 		/* make sure we contine with the next one in the chain */
							(*handle_data)(NULL,c_count,IO_REMOVE);	/* !!!!!!!!!tell the world that it's gone, handle_data should take care of waiting packets too */
							c = conn_del(c,c_count->id); 		/* make sure the main list has it removed and delete it forever */
							c_count = c_new;
							continue;
						}
					}

					buffer[nbytes] = '\0';
					c_new = (*handle_data)(buffer,c_count,IO_NORMAL);
					if((int)c_new == -1) /* special remove flag */
					{
						DBG_VAL("droppingconnection",c_count->name);
						FD_CLR(c_count->id, &active_fd_set); 	/* remove it from the select */
						c_new = c_count->next; 			/* make sure we contine with the next one in the chain */
						c = conn_del(c,c_count->id); 		/* make sure the main list has it removed and delete it forever */
						c_count = c_new;
						continue;
					}else{
						c = conn_append(c, c_new);
					}
				}
			}
			if(c_count->id > maxfd)
				maxfd = c_count->id;
			FD_SET(c_count->id, &active_fd_set);
			c_count = c_count->next;
		}

		/*
		 * Loop through all of the connections to write
		 */
		c_count = c;
		while(c_count != NULL)
		{
			if(c_count->packets != NULL)
			{
				DBG_VAL("writing_data_to",c_count->name);
				nbytes = write(
					c_count->id,
					c_count->packets->data + c_count->packets->sent,
					strlen(c_count->packets->data) - c_count->packets->sent);
				if(nbytes < 0)
				{ /* connection broken, return it to the select to let the main loop remove it */
					c_count = c_count->next;
					continue;
				}
				h_err(nbytes,"write");
				if((nbytes + c_count->packets->sent) == strlen(c_count->packets->data))
				{
					/* this packets empty, crush and move on */
					p = c_count->packets->next;
					free_packet(c_count->packets);
					c_count->packets = p;
					if(c_count->packets == NULL && FD_ISSET(c_count->id,&write_fd_set))
					{
						/* there's nothing left to write, dont wait anymore */
						FD_CLR(c_count->id,&write_fd_set);
						c_count = c_count->next;
						continue;
					}
				}else{
					c_count->packets->sent += nbytes;
				}
				/* if theres still stuff waiting, come back and get it later */
				if(c_count->packets != NULL)
					FD_SET(c_count->id,&write_fd_set);
			}
			c_count = c_count->next;
		}
	}

	/* shutdown */

	c_count = c;
	while(c_count != NULL)
	{
		close(c_count->id);
		c_count = c_count->next;
	}
	DBG_MSG("EXITING, sockets closed");
}


void syslog(int level, char *from, char *message)
{
	static time_t t;
	char *tstr, *tmpctime;

	if(_jabberlib_logger != NULL)
	{
		(*_jabberlib_logger)(level, from, message);
		return;
	}
	if(level >= _jabberlib_dbg)
	{
		time(&t);
		tstr = strgrow(NULL,"LIB_",0,0);
		if(level == SYSLOG_DEBUG)
			tstr = strgrow(tstr,"DEBUG: [",1,0);
		else
			tstr = strgrow(tstr,"ERROR: [",1,0);
		tmpctime = strdup(ctime(&t));
		tmpctime[strlen(tmpctime) - 1] = '\0';
		tstr = strgrow(tstr,tmpctime,1,1);
		tstr = strgrow(tstr,"] ",1,0);
		tstr = strgrow(tstr,from,1,0);
		tstr = strgrow(tstr," \"",1,0);
		if(message != NULL)
			tstr = strgrow(tstr,message,1,0);
		tstr = strgrow(tstr,"\" ",1,0);
		printf("%s\n",tstr);
		free(tstr);
	}
}


void syslog_debug(int level, char *from, char *name, char *val)
{
	char *tstr;

	tstr = strgrow(NULL,name,0,0);
	tstr = strgrow(tstr,"=",1,0);
	if(val != NULL)
		tstr = strgrow(tstr,val,1,0);
	else
		tstr = strgrow(tstr,"(NULL)",1,0);
	syslog(level, from, tstr);
	free(tstr);
}

