#include "common.h"

/* jpair.c
 *
 * Contains functions to manage the simple common linked list, jpair.
 * It's not designed to handle large or dynamic arrays, only small(<10) static ones(build, use, destroy).
 *
 */


/* concatenates two jpairs and returns pointer to result
 */
jpair *jpair_append(jpair *j1, jpair *j2)
{
	jpair *j_count = j1;

	if(j1 == NULL)
		return j2;
	if(j2 == NULL)
		return j1;
	while(j_count->next != NULL)
		j_count = j_count->next;
	j_count->next = j2;
	return j1;
}


/* creates new item in list from args
 * IMPORTANT, with jpair_add the key and val must be malloc'd memory!
 */
jpair *jpair_add(jpair *p, char *key, char *val)
{
	jpair *p_new;

	p_new = malloc(sizeof(jpair));
	p_new->val = val;
	p_new->key = key;
	p_new->next = NULL;

	if(p == NULL)
		return p_new;
	p_new->next = p;
	return p_new;
}


/* creates new item in the list
 * makes copies of the args, mallocs new memory
 */
jpair *jpair_new(jpair *p, char *key, char *val, int flag)
{
	jpair *p_new;

	p_new = malloc(sizeof(jpair));
	if(key == NULL)
		p_new->key = NULL;
	else
		p_new->key = strdup(key);
	if(val == NULL)
		p_new->val = NULL;
	else
		p_new->val = strdup(val);
	p_new->flag = flag;
	p_new->next = NULL;

	if(p == NULL)
		return p_new;
	p_new->next = p;
	return p_new;
}


void free_jpair(jpair *p)
{
	jpair *next;

	next = p;

	while(next != NULL)
	{
		p = next;
		next = p->next;
		free(p->key);
		free(p->val);
		free(p);
	}
}


/* creates a duplicate of the jpair
 * returns pointer to new duplicate
 */
jpair *jpair_dup(jpair *p)
{
	jpair *new = NULL;

	while(p != NULL)
	{
		new = jpair_new(new,p->key,p->val,p->flag);
		p = p->next;
	}
	return new;
}


/* locates the first matching entry and returns pointer to it
 */
jpair *jpair_get(jpair *p, char *key)
{
	while(p != NULL)
	{
		if(p->key != NULL && key != NULL && strcmp(p->key,key) == 0)
			return p;
		p = p->next;
	}
	return NULL;
}


/* locates the first matching entry and returns pointer to it's value
 */
char *jpair_getval(jpair *p, char *key)
{
	while(p != NULL)
	{
		if(p->key != NULL && key != NULL && strcmp(p->key,key) == 0)
			return p->val;
		p = p->next;
	}
	return NULL;
}


/* locates the first matching entry and returns a pointer to a new copy of the value
 */
char *jpair_getvalnew(jpair *p, char *key)
{
	char *val;

	val = jpair_getval(p, key);
	if(val != NULL)
		return strdup(val);
	return NULL;
}


/* locates the first matching entry and returns it's flag
 */
int jpair_getflag(jpair *p, char *key)
{
	while(p != NULL)
	{
		if(p->key != NULL && key != NULL && strcmp(p->key,key) == 0)
			return p->flag;
		p = p->next;
	}
	return 0;
}


/* converts the standard argument array into a jpair, very handy
 */
jpair *arg2jpair(int argc, char **argv)
{
	jpair *ret = NULL;
	int i;
	int first = 0;

	for(i = 1; i < argc; i++)
	{
		if(*argv[i] == '-')
		{
			ret = jpair_add(ret,strdup(argv[i]),NULL);
			first = 1;
		}else{
			if(first)
			{
				first = 0;
				ret->val = strdup(argv[i]);
			}else{
				ret = jpair_add(ret,strdup(argv[i]),NULL);
			}
		}
	}
	return ret;
}


/* converts Expat's attribute structure into a jpair
 */
jpair *atts2jpair(char **atts)
{
	jpair *ret = NULL;
	int i = 0;
	int name = 1;

	while(atts != NULL && atts[i] != '\0')
	{
		if(name)
		{
			ret = jpair_add(ret,strdup(atts[i]),NULL);
			name = 0;
		}else{
			ret->val = strdup(atts[i]);
			name = 1;
		}
		i++;
	}
	return ret;
}


/* converts a chain of tags into a jpair
 * example:
 *	<tag>value</tag><a>bcd</a>
 * parsed into xpt becomes a jpair with tag=value and a=bcd
 * mostly just a utility funciton
 */
jpair *xpt2jpair(xpt *x)
{
	jpair *ret = NULL;
	char *key, *val;
	xpt *next = NULL;

	while(x != NULL)
	{
		if(x->type == XPT_TAG)
		{
			key = strdup(x->name);
			val = NULL;
			next = x->children;
			while(next != NULL)
			{
				if(next->type == XPT_CDATA)
					val = strgrow(val, next->value,1,0);
				next = next->next;
			}
			ret = jpair_add(ret, key, val);
		}
		x = x->next;
	}

	return ret;
}


/* for debugging, just prints out a jpair
 */
void jpair_print(jpair *p)
{
	printf("===PAIR===[");
	while(p != NULL)
	{
		printf("%s[%s] ",p->key, p->val);
		p = p->next;
	}
	printf("]===/PAIR===\n");
}
