
#include "common.h"


/* Define common escape sequences */
#define C_AMP "&amp;"
#define C_APOS "&apos;"
#define C_QUOT "&quot;"
#define C_LT "&lt;"
#define C_GT "&gt;"
/* Define common escape sequence string length */
#define C_AMP_LEN   5
#define C_APOS_LEN  6
#define C_QUOT_LEN  6
#define C_LT_LEN    4
#define C_GT_LEN    4
/* Give each escape sequence a unique id */
#define C_AMP_ID  0
#define C_APOS_ID 1
#define C_QUOT_ID 2
#define C_LT_ID   3
#define C_GT_ID   4
/* Define how many escape sequences there are */
#define ESCAPE_COUNT 5

/* strescape
   Takes an input string and ensures that proper XML escape sequences
   are used for the funky chars (&,',",<,>)
*/
char* strescape(const char* s)
{
     int i,j;		      
     int count_table[ESCAPE_COUNT];
     int count_tbl_sum = 0;
     char* result;
     int   resultsz = 0;

     /* Initialize count_table */
     for (i = 0; i < ESCAPE_COUNT; i++)
	  count_table[i] = 0;
     /* Ensure string to check is not NULL */
     if (s == NULL)
	  return NULL;
     /* Loop through the string, counting occurrences
	of escape values */
     for (i = 0; i < strlen(s); i++) 
     {
	  switch (s[i]) 
	  {
	  case '&': count_table[C_AMP_ID]++;  count_tbl_sum++;break;
	  case '\'':count_table[C_APOS_ID]++; count_tbl_sum++;break;
	  case '\"':count_table[C_QUOT_ID]++; count_tbl_sum++;break;
	  case '<': count_table[C_LT_ID]++;   count_tbl_sum++;break;
	  case '>': count_table[C_GT_ID]++;   count_tbl_sum++;break;
	  }
     }
     /* Allocate a buffer which will be the final size of the string,
	using our count_table to help determine the size */
     resultsz = ((count_table[C_AMP_ID] * C_AMP_LEN) +
		 (count_table[C_APOS_ID] * C_APOS_LEN) +
		 (count_table[C_QUOT_ID] * C_QUOT_LEN) +
		 (count_table[C_LT_ID] * C_LT_LEN) +
		 (count_table[C_GT_ID] * C_GT_LEN) +
		 strlen(s) - count_tbl_sum);
     /* Allocate the actual buffer, adjust for null terminator */
     result = (char*)malloc(resultsz + 1);
     /* Now, loop through the string once more, copying and replacing
	characters accordingly... */
     j = 0;
     for (i = 0; i < strlen(s); i++)
     {
	  switch (s[i])
	  {
	  case '&' : memcpy(&result[j], C_AMP, C_AMP_LEN);  j+= C_AMP_LEN;  break;
	  case '\'': memcpy(&result[j], C_APOS, C_APOS_LEN);j+= C_APOS_LEN; break;
	  case '\"': memcpy(&result[j], C_QUOT, C_QUOT_LEN);j+= C_QUOT_LEN; break;
	  case '<' : memcpy(&result[j], C_LT, C_LT_LEN);    j+= C_LT_LEN;   break;
	  case '>' : memcpy(&result[j], C_GT, C_GT_LEN);    j+= C_GT_LEN;   break;
	  default  : result[j++] = s[i];
	  }
     }
     /* Ensure new string is null terminated */
     result[resultsz] = '\0';
     return result;
}

char *strdescape(char *s)
{
	char *str1, *str2, *str3, *str4, *ret;

	if(s == NULL)
		return NULL;

	if(	strstr(s,"&gt;") != NULL ||
		strstr(s,"&lt;") != NULL ||
		strstr(s,"&quot;") != NULL ||
		strstr(s,"&apos;") != NULL ||
		strstr(s,"&amp;") != NULL)
	{
		str1 = strrepl(s,"&gt;",">");
		str2 = strrepl(str1,"&lt;","<");
		str3 = strrepl(str2,"&quot;","\"");
		str4 = strrepl(str3,"&apos;","'");
		ret = strrepl(str4,"&amp;","&");
		free(str1);
		free(str2);
		free(str3);
		free(str4);
	}else{
		ret = strdup(s);
	}
	return ret;
}

char *substr(char *s, char *start, char *end)
{
	char *ret, *sp, *ep;

	sp = strstr(s,start);
	ep = strstr(s,end);
	if(sp == NULL || ep == NULL)
		return NULL;

	sp += strlen(start);
	ret = subpstr(sp,ep);

	return ret;
}

char *subpstr(char *s, char *end)
{
	char *sub;
	int len;

	len = (int)end - (int)s;
	sub = malloc(len + 1);
	if(len > 0)
		strncpy(sub,s,len);
	sub[len] = '\0';

	return sub;
}

int strstrcount(char *s, char *sub)
{
	int i = 0;
	char *curr = s;

	curr = strstr(curr,sub);
	while(curr != NULL)
	{
		i++;
		curr++;
		curr = strstr(curr,sub);
	}

	return i;
}

char *strrepl(char *s, char *oldstr, char *newstr)
{
	int i;
	char *ret, *lastcurr, *subs;
	char *curr = s;
	i = strstrcount(s,oldstr);
	ret = malloc(strlen(s) + 1 + ((strlen(newstr) - strlen(oldstr)) * i) );
	ret[0] = '\0';

	if(strstr(s,oldstr) == NULL)
	{
		strcpy(ret,s);
		return ret;
	}

	while(1)
	{
		lastcurr = curr;
		curr = strstr(curr,oldstr);
		if(curr == NULL)
		{
			strcat(ret,lastcurr);
			break;
		}
		else
		{
			subs = subpstr(lastcurr,curr);
			strcat(ret,subs);
			strcat(ret,newstr);
			free(subs);
			curr += strlen(oldstr);
		}
	}
	return ret;
}


/* Very scary function I use everywhere...
 * It is simply a growth buffer, mallocs a string in 100 byte blocks.
 * Not sure it really helps anything, but it works happily and is convenient
 */

char *strgrow(char *oldstr, char *newstr, int old_free, int new_free)
{
	char *ret = oldstr;
	int i,ol,nl,left;

	if(newstr == NULL)
		return ret;

	nl = strlen(newstr) + 1;
	if(oldstr != NULL)
	{
		ol = strlen(oldstr);
	}else{
		ol = 0;
	}
	left = 100 - (ol % 100);
	if(left < nl || oldstr == NULL)
	{
		i = (((ol + nl)/100)*100) + 100;
		ret = malloc(i);
		ret[0] = '\0';
		if(oldstr != NULL)
			strcpy(ret,oldstr);
		if(old_free)
			free(oldstr);
	}

	strcat(ret,newstr);
	if(new_free)
		free(newstr);
	return ret;
}


