#include "common.h"

/* xpt.c
 * 
 * This is where the XML parsing goes on, via callbacks from Expat.
 * It contains some wrapper functions that hide the complexity,
 * and store everything in the xpt structure.  There are also some
 * utility functions for maintaining the xpt structures
 */


/* Expat calls this when it has found the start of a new tag: <tag>
 */
void expat_startElement(void *userData, const char *name, const char **atts)
{
	int i;
	xptpool *xp = userData;
	xpt *x;

	x = xpt_new_tag(xp->buff, (char *)name, (char **)atts);
	x->status = XPT_INCOMPLETE;
	xp->buff = x;
}


/* Expat calls this when it has found the closing tag: </tag>
 */
void expat_endElement(void *userData, const char *name)
{
	xptpool *xp = userData;

	xp->buff->status = XPT_COMPLETE;

	/* reset the current xpt up a level, unless we're at the top already */
	if(xp->buff->parent != NULL)
		xp->buff = xp->buff->parent;

}


/* Expat calls this whenever it finds any strings outside of a tag or between a tag
 * Note: <tag>VALUE</tag> is not guaranteed to be called as "VALUE" but might end up as two calls "VA" and "LUE"!
 */
void expat_charData(void *userData, const char *s, int len)
{
	xptpool *xp = userData;
	xpt *x;
        char *str;

        str = malloc(len + 1);

	memcpy(str,s,len);
	str[len]='\0';

	/* throw it away if we're not in a tag */
	if(xp->buff != NULL)
		x = xpt_new_cdata(xp->buff,str);

	free(str);
}


/*
 * expat_*_packet
 *    These are only called for the opening "root" tag in a packetized decoding
 *
 */
void expat_startElement_packet(void *userData, const char *name, const char **atts)
{
	int i;
	xptpool *xp = userData;

	if(name != NULL)
	{
		xp->name = (char *)strdup(name);
	}
	xp->attributes = atts2jpair((char **)atts);

	/* start processing normally now, but with our wrapper to sort them out */
	XML_SetElementHandler(xp->parser, expat_startElement, expat_endElement_packet);
	XML_SetCharacterDataHandler(xp->parser, expat_charData);
}


/*
 * SPECIAL: This is a wrapper around the real closing function, it breaks the results into packets
 */
void expat_endElement_packet(void *userData, const char *name)
{
	xptpool *xp = userData;
	xpt *next;

	/* this means we've recieved the closing tag for the stream */
	if(xp->buff == NULL)
	{
		xp->flag = XPT_POOLFLAG_CLOSE;
		return;
	}

	expat_endElement(userData, name);

	/* When a packet has been processed fully, move it into the packets array */
	if(xp->buff->parent == NULL && xp->buff->status == XPT_COMPLETE)
	{
		xp->buff->next = NULL;
		if(xp->packets == NULL)
		{
			xp->packets = xp->buff;
		}else{
			/* make sure we're appending to the end */
			next = xp->packets;
			while(next != NULL)
			{
				if(next->next == NULL)
				{
					next->next = xp->buff;
					break;
				}
				next = next->next;
			}
		}
		xp->buff = NULL;
	}
}


/*
 * xpt_packet_*
 *    To process an xml "stream" into packets, call xpt_packet_new to create the right pool,
 *    and as data comes into the stream, call xpt_packet_data.  After that, check the pool->packets
 *    for a linked list of packets found in the stream.
 *
 */
xptpool *xpt_packet_new()
{
	xptpool *xp;

	xp = xpt_newpool();
	XML_SetUserData(xp->parser, xp);
	XML_SetElementHandler(xp->parser, expat_startElement_packet, expat_endElement_packet);

	return xp;
}


void xpt_packet_data(xptpool *xp, char *str, int len)
{
	if(str != NULL && len > 0)
	{
		if(!XML_Parse(xp->parser, str, len, 0)) /* !! last value in call, isFinal, need to set for </jabber> or something? */
		{
			xp->flag = XPT_POOLFLAG_ERROR;
			DBG_MSG("EXPAT: XML processing error");
		}
	}
}


/* Pass a file name and every attempt will be made to parse it's contents as XML and return an xpt containing the results
 * NULL will be returned on any failures.
 */
xpt *xpt_file(char *fname)
{
	char buf[BUFSIZ];
	xptpool *xp;
	int done, f, len;
	xpt *ret;

	if(NULL == fname)
		return NULL;

	DBG_VAL("parsing_file",fname);

	f = open(fname,O_RDONLY);
	if(f < 0)
		return NULL;

	xp = xpt_newpool();
	XML_SetUserData(xp->parser, xp);
	XML_SetElementHandler(xp->parser, expat_startElement, expat_endElement);
	XML_SetCharacterDataHandler(xp->parser, expat_charData);

	do{
		len = read(f, buf, sizeof(buf)); 
		done = len < sizeof(buf);
		if(!XML_Parse(xp->parser, buf, len, done))
		{
			DBG_VAL("parsing_file_xml_error",(char *)XML_ErrorString(XML_GetErrorCode(xp->parser)));
			return NULL;
		}

	}while(!done);

	ret = xp->buff;
	free_xptpool(xp);
	return ret;
}


/* Parses a string as an XML document, returns NULL if anything fails
 */
xpt *xpt_str(char *str)
{
	xpt *ret;
	xptpool *xp = NULL;

	if(NULL == str)
		return NULL;

	xp = xpt_newpool();

	XML_SetUserData(xp->parser, xp);
	XML_SetElementHandler(xp->parser, expat_startElement, expat_endElement);
	XML_SetCharacterDataHandler(xp->parser, expat_charData);

	if(!XML_Parse(xp->parser, str, strlen(str), 1))
		return NULL;

	ret = xp->buff;
	free_xptpool(xp);
	return ret;
}


/* Internal function used to create a new xpt and initialize it
 */
xpt *xpt_new(xpt *parent)
{
	xpt *x, *next, *prev;

	x = (xpt *)malloc(sizeof(xpt));

	/* initialize */
	x->type = 0;
	x->name = NULL;
	x->value = NULL;
	x->children = NULL;
	x->attributes = NULL;
	x->parent = parent;

	if(parent != NULL)
	{
		/* insert into parent's list of children */
		x->next = NULL;
		if(parent->children == NULL)
		{
			parent->children = x;
		}else{
			/* make sure we're appending to the end */
			next = parent->children;
			while(next != NULL)
			{
				if(next->next == NULL)
				{
					next->next = x;
					break;
				}
				next = next->next;
			}
		}
	}else{
		x->next = NULL;
	}

	return x;

	return NULL;
}


/* Inserts a new tag into the parents xpt and copies values into it
 */
xpt *xpt_new_tag(xpt *parent, char *name, char **atts)
{
	xpt *x;

	x = xpt_new(parent);
	if(x == NULL)
		return x;

	x->type = XPT_TAG;
	if(name != NULL)
	{
		x->name = (char *)strdup(name);
	}
	x->attributes = atts2jpair(atts);
	return x;
}


/* Adds in a string to the xpt
 */
xpt *xpt_new_cdata(xpt *parent, char *data)
{
	xpt *x;

	x = xpt_new(parent);
	if(x == NULL)
		return x;

	x->type = XPT_CDATA;
	if(data != NULL)
	{
		x->value = (char *)strdup(data);
		/* !!! DO: strgrow parent->value with data */
	}
	return x;
}


/* creates and initializes a new xpt pool
 */
xptpool *xpt_newpool()
{
	xptpool *xp;

	DBG_MSG("Creating new XML Parser pool");

	xp = malloc(sizeof(xptpool));
	xp->name = NULL;
	xp->attributes = NULL;
	xp->packets = NULL;
	xp->buff = NULL;
	xp->flag = XPT_POOLFLAG_NONE;
	xp->parser = XML_ParserCreate(NULL);

	return xp;
}

void free_xpt(xpt *x)
{
	xpt *next, *old;

	if(x != NULL)
	{
		free(x->name);
		free(x->value);
		free_jpair(x->attributes);
		next = x->children;
		while(next != NULL)
		{
			old = next;
			next = next->next;
			free_xpt(old);
		}
		free(x);
	}
}

void free_xptpool(xptpool *xp)
{
	if(xp != NULL)
	{
		/* xp->buff is long-term, free everything else */
		free(xp->name);
		free_jpair(xp->attributes);
		free_xpt(xp->packets);
		XML_ParserFree(xp->parser);
	}
}
 

/* Translates an xpt structure into a string again, let's us see the parsed XML as a string
 */
char *xpt2str(xpt *x)
{
	jpair *atts;
	xpt *next;
	char *ret = NULL;
	char *buf = NULL;

	if(x->type == XPT_CDATA)
	{
		/* kludge, so a string that is already XML can be passed through unescaped */
		if(x->name != NULL && strcmp(x->name, "NOESCAPE") == 0)
			ret = strgrow(ret,x->value,1,0);
		else
			ret = strgrow(ret,strescape(x->value),1,1);
	}else{
		ret = strgrow(ret,"<",1,0);
		ret = strgrow(ret,x->name,1,0);
		atts = x->attributes;
		while(atts != NULL)
		{
			ret = strgrow(ret," ",1,0);
			ret = strgrow(ret,atts->key,1,0);
			ret = strgrow(ret,"='",1,0);
			ret = strgrow(ret,strescape(atts->val),1,1);
			ret = strgrow(ret,"'",1,0);
			atts = atts->next;
		}

		next = x->children;
		buf = NULL;
		while(next != NULL)
		{
			buf = strgrow(buf,xpt2str(next),1,1);
			next = next->next;
		}
		if(buf == NULL)
		{
			ret = strgrow(ret,"/>",1,0);
		}else{
			ret = strgrow(ret,">",1,0);
			ret = strgrow(ret,buf,1,1);
			ret = strgrow(ret,"</",1,0);
			ret = strgrow(ret,x->name,1,0);
			ret = strgrow(ret,">",1,0);
		}
	}

	return ret;
}


/* just comes in handy to combine broken CDATA sections together */
char *xpt2str_chain(xpt *x)
{
	char *ret = NULL;
	while(x != NULL)
	{
		if(x->type == XPT_CDATA)
			ret = strgrow(ret,x->value,1,0);
		x = x->next;
	}
	return ret;
}


/* Debug function, just simply prints out an xpt
 */
void xpt_print(xpt *x, int detail)
{
	jpair *atts;

	while(x != NULL)
	{
		if(x->type == XPT_CDATA)
		{
			if(detail >= 1)
				printf("%s",x->value);
		}else{
			printf("<%s",x->name);
			atts = x->attributes;
			while(atts != NULL && detail >= 2)
			{
				printf(" %s='%s'",atts->key,atts->val);
				atts = atts->next;
			}
			printf(">");
			xpt_print(x->children, detail);
			printf("</%s>",x->name);
		}
		x = x->next;
	}
}


