#include "jabber.h"

/*   The intention here is that the modules array contains
 * pointers to functions in each module, and modules_initialize
 * loops through and calls each of them when ready.
 */

typedef void (*jfptr_init)(xpt *config);

void mod_group_init(xpt *config);
void mod_test_init(xpt *config);
void mod_basic_init(xpt *config);

jfptr_init modules[] = {
	mod_group_init,
	mod_test_init,
	mod_basic_init,
	NULL
};

void modules_initialize(xpt *config)
{
	jfptr_init *m;

	DBG_MSG("Module initialization");

	/* do any dynamic loading of modules here */

	for(m = modules; *m != NULL; m++)
	{
		(*m)(config);
	}
}

/* fetches the right handler from the list of them */
handler *modules_gethandler_id(int phase, char *id)
{
	handler *next;

	DBG_VAL("get_handler",id);

	next = etc.handlers;
	while(next != NULL)
	{
		if(next->type == phase && id != NULL && strcmp(next->name, id) == 0)
		{
			return next;
		}
		next = next->next;
	}

	return NULL;
}

char *modules_getgroup(char *user)
{
	char *ret = NULL;
	handler *next;
	static jpair *cache;

	DBG_VAL("get_group_user", user);

	/* check for cached value */
	if(jpair_getval(cache, user) != NULL)
		return jpair_getvalnew(cache, user);

	next = etc.handlers;
	while(next != NULL)
	{
		if(next->type == API_GROUP)
		{
			ret = (next->h_group)(user);
			if(ret != NULL)
			{
				cache = jpair_new(cache, user, ret, 0);
				return ret;
			}
		}
		next = next->next;
	}
	return NULL;
}

handler *modules_gethandler_group(int phase, char *group)
{
	char *module, *maybe;
	xpt *next;
	jpair *par;
	handler *ret;

	/*
	 * matching user's group ID to a module ID:
	 *	if a group's phase is configured to a module, use that: <group id="test"><auth>mod_auth_dbm</auth></group>
	 *	if a group's default module is set, use that: <group id="test" default="mod_mysql"/>
	 *	if nothing else, module ID == group ID, such as nothing is configured, do it manually
	 */

	DBG_VAL("get_handler_group",group);

	module = NULL;

	if(etc.modconfig != NULL)
	{
		next = etc.modconfig->children;
		while(next != NULL)
		{
			if(next->type == XPT_TAG && strcmp(next->name, "group") == 0 && strcmp(jpair_getval(next->attributes,"id"), group) == 0)
			{
				module = jpair_getvalnew(next->attributes, "default");
				par = xpt2jpair(next->children);
				switch(phase)
				{
					case API_AUTH:
						maybe = jpair_getval(par,"auth");
						break;
					case API_QUERY:
						maybe = jpair_getval(par,"info");
						break;
					case API_STATUSIN:
						maybe = jpair_getval(par,"roster");
						break;
					case API_STATUSOUT:
						maybe = jpair_getval(par,"roster");
						break;
					case API_ROSTER:
						maybe = jpair_getval(par,"roster");
						break;
					case API_LOGIN:
						maybe = jpair_getval(par,"messages");
						break;
					case API_LOGOUT:
						maybe = jpair_getval(par,"messages");
						break;
					case API_MESSAGEIN:
						maybe = jpair_getval(par,"messages");
						break;
					case API_MESSAGEOUT:
						maybe = jpair_getval(par,"messages");
						break;
				}
				if(maybe != NULL)
					module = strdup(maybe);
				free_jpair(par);
			}
			next = next->next;
		}
	}

	if(module == NULL)
		module = strdup(group);

	DBG_VAL("handler_in_module",module);

	ret = modules_gethandler_id(phase, module);
	free(module);
	return ret;
}

handler *modules_gethandler_user(int phase, char *user)
{
	char *group;
	handler *ret;

	group = modules_getgroup(user);
	ret = modules_gethandler_group(phase, group);
	free(group);
	return ret;
}


/*   Each module init function, when called, would 
 * check the config xml structure for it's configuration data
 * and then run the below functions to define it's callbacks for
 * the particular API phases.
 */


/* stores an auth handler function to try */
void module_set_auth(char *id, int (*authhandler)(char *user, char *pass))
{
	handler *auth;

	DBG_VAL("module_set_auth",id);

	auth = malloc(sizeof(handler));

	auth->name = strdup(id);
	auth->h_auth = authhandler;
	auth->type = API_AUTH;
	auth->next = etc.handlers;
	etc.handlers = auth;
}

/* stores a group handler function to try */
void module_set_group(char *id, char *(*grouphandler)(char *user))
{
	handler *group;

	DBG_VAL("module_set_group",id);

	group = malloc(sizeof(handler));
	group->name = strdup(id);
	group->h_group = grouphandler;
	group->type = API_GROUP;
	group->next = etc.handlers;
	etc.handlers = group;
}

/* stores a pair of login/logout handlers for a module id */
void module_set_log(char *id, void (*log_in)(char *user, char *nick), void (*log_out)(char *user, char *nick))
{
	handler *li, *lo;

	DBG_VAL("module_set_log",id);

	if(log_in != NULL)
	{
		li = malloc(sizeof(handler));
		li->name = strdup(id);
		li->h_log = log_in;
		li->type = API_LOGIN;
		li->next = etc.handlers;
		etc.handlers = li;
	}

	if(log_out != NULL)
	{
		lo = malloc(sizeof(handler));
		lo->name = strdup(id);
		lo->h_log = log_out;
		lo->type = API_LOGOUT;
		lo->next = etc.handlers;
		etc.handlers = lo;
	}
}

/* stores a pair of message handlers for a module id */
void module_set_message(
	char *id,
	int (*message_in)(char *localuser, char *localnick, char *otheruser, char *othernick, char *subject, char *say, char *thread, char *ext, int priority, char *type),
	int (*message_out)(char *localuser, char *localnick, char *otheruser, char *othernick, char *subject, char *say, char *thread, char *ext, int priority, char *type))
{
	handler *mi, *mo;

	DBG_VAL("module_set_message",id);

	if(message_in != NULL)
	{
		mi = malloc(sizeof(handler));
		mi->name = strdup(id);
		mi->h_message = message_in;
		mi->type = API_MESSAGEIN;
		mi->next = etc.handlers;
		etc.handlers = mi;
	}

	if(message_out != NULL)
	{
		mo = malloc(sizeof(handler));
		mo->name = strdup(id);
		mo->h_message = message_out;
		mo->type = API_MESSAGEOUT;
		mo->next = etc.handlers;
		etc.handlers = mo;
	}
}


/* stores a pair of login/logout handlers for a module id */
void module_set_status(char *id, int (*status_in)(char *to, char *user, char *nick, int type, char *say, char *icon, char *ext, int priority), jpair *(*status_out)(char *user, char *nick, int type, char *say, char *icon, char *ext, int priority))
{
	handler *si, *so;

	DBG_VAL("module_set_status",id);

	if(status_in != NULL)
	{
		si = malloc(sizeof(handler));
		si->name = strdup(id);
		si->h_status_in = status_in;
		si->type = API_STATUSIN;
		si->next = etc.handlers;
		etc.handlers = si;
	}

	if(status_out != NULL)
	{
		so = malloc(sizeof(handler));
		so->name = strdup(id);
		so->h_status_out = status_out;
		so->type = API_STATUSOUT;
		so->next = etc.handlers;
		etc.handlers = so;
	}

}

/* stores roster handlers for a module id */
void module_set_roster(char *id, jpair *(*roster_handler)(char *localuser, int type, char *group, char *user))
{
	handler *r;

	DBG_VAL("module_set_roster",id);

	if(roster_handler != NULL)
	{
		r = malloc(sizeof(handler));
		r->name = strdup(id);
		r->h_roster = roster_handler;
		r->type = API_ROSTER;
		r->next = etc.handlers;
		etc.handlers = r;
	}
}

/* stores query handlers for a module id */
void module_set_query(char *id, void *queryhandler)
{

}

